/* Copyright 2020-2021 Ethan Halsall
    This file is part of icecast-metadata-js.

    icecast-metadata-js free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    icecast-metadata-js distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>
*/

const noOp = () => {};

export default class IcecastMetadataQueue {
  /**
   * @description Schedules updates up to the millisecond for Icecast Metadata from the response body of an Icecast stream mountpoint
   * @description The accuracy of metadata updates is a direct relationship of the icyMetaInt
   * @param {Object} IcecastMetadataQueue constructor parameter
   * @param {number} [IcecastMetadataQueue.icyBr] Bitrate of audio stream used to increase accuracy when to updating metadata
   * @param {onMetadataUpdate} [IcecastMetadataQueue.onMetadataUpdate] Callback executed when metadata is scheduled to update
   * @param {onMetadataEnqueue} [IcecastMetadataQueue.onMetadataEnqueue] Callback executed when metadata is enqueued
   * @param {boolean} [IcecastMetadataQueue.paused] Set to true to start the queue in a paused mode
   *
   * @callback onMetadataUpdate
   * @param {Object} metadata Object containing all metadata received.
   * @param {string} [metadata.StreamTitle] Title of the metadata update.
   * @param {string} [metadata.StreamUrl] Url (usually album art) of the metadata update.
   * @param {number} timestampOffset Total time buffered when the metadata was added
   * @param {number} timestamp Current time of the audio player when the metadata was added
   *
   * @callback onMetadataEnqueue
   * @param {Object} metadata Object containing all metadata received.
   * @param {string} [metadata.StreamTitle] Title of the metadata update.
   * @param {string} [metadata.StreamUrl] Url (usually album art) of the metadata update.
   * @param {number} timestampOffset Total time buffered when the metadata was added
   * @param {number} timestamp Current time of the audio player when the metadata was added
   *
   */
  constructor({
    icyBr,
    onMetadataUpdate = noOp,
    onMetadataEnqueue = noOp,
    paused = false,
  }) {
    this._icyBr = icyBr;
    this._onMetadataUpdate = onMetadataUpdate;
    this._onMetadataEnqueue = onMetadataEnqueue;
    this._initialPaused = paused;
    this._paused = paused;
    this._isInitialMetadata = true;
    this._pendingMetadata = [];
    this._metadataQueue = [];
  }

  /**
   * @description Returns the metadata queued for updates
   * @type {{metadata: string, time: number}[]} Queued metadata
   */
  get metadataQueue() {
    return this._metadataQueue.map(({ _timeoutId, ...rest }) => rest);
  }

  /**
   *
   * @param {object} metadata Metadata object returned from IcecastMetadataReader
   * @param {number} timestampOffset Total buffered audio in seconds
   * @param {number} [timestamp] Current time in the audio player
   */
  addMetadata({ metadata, stats }, timestampOffset, timestamp = 0) {
    /**
     * Metadata time is derived from the total number of stream bytes read
     * since the latest buffer input. The buffer offset should be the total
     * seconds of audio in the player buffer when the metadata was read.
     */
    timestamp += stats ? this.getTimeByBytes(stats.currentStreamPosition) : 0;

    const metadataPayload = {
      metadata,
      timestampOffset,
      timestamp,
    };

    this._paused
      ? this._pendingMetadata.push(metadataPayload)
      : this._enqueueMetadata(metadataPayload);
  }

  /**
   * @description Calculates audio stream length based on bitrate
   * @param {number} bytesRead Number of bytes
   * @type {number} Seconds
   */
  getTimeByBytes(bytesRead) {
    return this._icyBr ? bytesRead / (this._icyBr * 125) : 0;
  }

  /**
   * @description Starts the metadata queue if it was paused
   * @param {number} [timestamp] Current time of the audio player
   */
  startQueue(timestamp) {
    if (this._paused) {
      this._pendingMetadata.forEach((u) => {
        if (timestamp !== undefined) u.timestamp = timestamp;

        // remove any metadata that has already elapsed
        if (timestamp === undefined || u.timestampOffset >= u.timestamp)
          this._enqueueMetadata(u);
      });
      this._pendingMetadata = [];
      this._paused = false;
    }
  }

  /**
   * @description Clears all metadata updates and empties the queue
   */
  purgeMetadataQueue() {
    this._metadataQueue.forEach((i) => clearTimeout(i._timeoutId));
    this._metadataQueue = [];
    this._pendingMetadata = [];
    this._paused = this._initialPaused;
    this._isInitialMetadata = true;
  }

  _enqueueMetadata(payload) {
    this._metadataQueue.push(payload);

    this._onMetadataEnqueue(
      payload.metadata,
      payload.timestampOffset,
      payload.timestamp,
    );

    if (this._isInitialMetadata) {
      this._dequeueMetadata();
      this._isInitialMetadata = false;
    } else {
      payload._timeoutId = setTimeout(
        () => {
          this._dequeueMetadata();
        },
        (payload.timestampOffset - payload.timestamp) * 1000,
      ); // trigger timeout relative to play position
    }
  }

  _dequeueMetadata() {
    // purging the queue is an async operation, some timeouts may still be queued in the event loop _after_ the purge call has happened
    if (this._metadataQueue.length) {
      const { metadata, timestampOffset, timestamp } =
        this._metadataQueue.shift();
      this._onMetadataUpdate(metadata, timestampOffset, timestamp);
    }
  }
}
